<!doctype html><html lang="en"><head><base href="{{base_url}}"/><meta charset="utf-8"/><meta name="viewport" content="width=device-width,initial-scale=1"/><link rel="manifest" href="manifest.json"/><link rel="icon" type="image/png" href="favicon.png" class="taipy-favicon" data-url="{{favicon}}"/><link rel="apple-touch-icon" href="favicon.png"/><title>{{title}}</title><script>{%- if config -%}
    window.taipyConfig={{config|tojson|safe}};{% endif %}
    {%- if version -%}
    window.taipyVersion="{{version}}";
    {% endif %}</script>{%- if css_vars -%}<style>:root{
      {{css_vars|safe}}
    }</style>{% endif %}<script defer="defer" src="taipy-gui-deps.dll.js?2e5e009c67dfd8a6d450"></script><script defer="defer" src="taipy-gui.js?2e5e009c67dfd8a6d450"></script><script defer="defer" src="taipy-gui-dom.js?2e5e009c67dfd8a6d450"></script></head><body>{% for script in scripts %}<script defer="defer" src="{{script}}"></script>{% endfor %} {% for css in styles %}<link rel="stylesheet" href="{{css}}"/>{% endfor %}<noscript>You need to enable JavaScript to run this app.</noscript><div id="root" {%- if root_margin -%} style="margin:{{root_margin}};" {% endif %}></div>{%- if watermark -%}<span class="taipy-watermark" style="position:fixed;bottom:1em;opacity:0.4;color:gray;left:50%;transform:translateX(-50%);font-size:0.7em;">{{watermark}}</span>{% endif %}</body></html>