/*
 * Copyright 2021-2024 Avaiga Private Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */

/**************************************************************

                   TAIPY SLIDER

***************************************************************/

/*************************************************
              OVERRIDES / NORMALIZATION
**************************************************/

.taipy-slider {
    display: inline-flex;
    max-width: 100%;
    min-height: 48px;
    align-items: center;
    margin: 4px 0;
}
