/*
 * Copyright 2021-2024 Avaiga Private Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */

/**************************************************************

                   TAIPY CHART

***************************************************************/

/*************************************************
              OVERRIDES / NORMALIZATION
**************************************************/

.taipy-chart {
  border-radius: var(--border-radius);
  background-color: var(--color-paper);
  box-shadow: var(--box-shadow);
  overflow: hidden;
}

.card .taipy-chart {
  border-radius: 0;
  box-shadow: none;
}

/*************************************************
              UTILITY CLASSES
**************************************************/

/* .has-background: keep the original background */
.taipy-chart:not(.has-background) .main-svg {
  background: transparent !important;
}

.js-plotly-plot.full-screen {
    position: fixed !important;
    height: 99vh !important;
    width: 99vw !important;
    display: block !important;
    left: 0;
    top: 0;
    z-index: 99999;
    overflow: hidden;
    background-color: var(--color-background);
    box-shadow: 10px 5px 5px var(--color-contrast);
    transition: left 1s ease, width 1s ease, top 1s ease, height 1s ease;
}
